"""
Describes generation of the features.

In variable ``features_to_extract`` are defined the features to be created, sources of their data and the methods
used for their processing. Features are created for every type of data entity (feature entity) individually.
Feature/s is/are described by the dict with keys as follows:

* **c_name_feat** - list of strings - Names of the generated features (column`s name).
* **c_name** - list of strings - Names of the data columns, that are used for creation of the features.
* **f** - method - identifier of the method (method`s name), that is used for creation of the feature/s **c_name_feat**
"""

import datetime as dt
import numpy as np

from ..support_files import ColumnNamesRaw, ColumnNames, ColumnNamesRawFiveMin
from ..configs.sections_settings import SENSOR_TYPE


def _engineer_time(data):
    """
    Function for calculating of the features - day of week (1-7) and hour of day

    :param np.datetime64 data: representation of datetime
    :return: list - day-of-week, hour-of-day
    """
    dd = dt.datetime.utcfromtimestamp(data.item() / 1e9)
    return [dd.isoweekday(), dd.hour]


def _engineer_time_bins(data):
    """
    Function for calculating of the feature - bin of the time. Time of the day in minutes is binarized into bins of size 10.

    :param np.ndarray data: array of np.datetime64
    :return: np.array, one value of bin per item in input array
    """
    dd = map(lambda x: dt.datetime.utcfromtimestamp(x.item() / 1e9), data)
    minutes = map(lambda x: x.hour * 60 + x.minute, dd)
    return np.digitize(list(minutes), list(range(0, 1440, 10)))


def _engineer_days(data):
    """
    Function for calculating of the features - days and weekend

    :param np.datetime64 data:  representation of datetime
    :return: list - weekend, monday, tuesday, wednesday, thursday, friday
    """
    dd = dt.datetime.utcfromtimestamp(data.item() / 1e9)

    l = [0] * 5
    if dd.weekday() < 5:
        l[dd.weekday()] = 1
        l = [0] + l
    else:
        l = [1] + l
    return l


# Here the features for extraction are defined...
features_to_extract = {
    SENSOR_TYPE.SENSOR_DET1: [
        {'c_name_feat': [ColumnNames.FEAT_det1_VELOCITY], 'c_name': [ColumnNames.AVG_VELOCITY], 'f': None},
        {'c_name_feat': [ColumnNames.FEAT_det1_OCCUPANCY], 'c_name': [ColumnNames.TOTAL_OCCUPANCY], 'f': None},
        {'c_name_feat': [ColumnNames.FEAT_det1_COUNT], 'c_name': [ColumnNames.NORM_COUNT], 'f': None}],
    SENSOR_TYPE.SENSOR_DET2:
        [
        {'c_name_feat': [ColumnNames.FEAT_det2_COUNT], 'c_name': [ColumnNamesRawFiveMin.LP_COUNT], 'f': None}],
}

features_to_extract_bck_tt =[
        {'c_name_feat': [ColumnNames.FEAT_TT_BCK], 'c_name': [ColumnNamesRaw.TT_CALCULATED], 'f': None}]

features_to_extract_time = [
    {'c_name_feat': [ColumnNames.FEAT_DAY, ColumnNames.FEAT_TIME], 'c_name': [ColumnNames.CALC_TIME], 'f': lambda x: map(_engineer_time, x)},
    {'c_name_feat': [ColumnNames.FEAT_WEEKEND,
                     ColumnNames.FEAT_MONDAY,
                     ColumnNames.FEAT_TUESDAY,
                     ColumnNames.FEAT_WEDNESDAY,
                     ColumnNames.FEAT_THURSDAY,
                     ColumnNames.FEAT_FRIDAY
                     ], 'c_name': [ColumnNamesRaw.CALC_TIME], 'f': lambda x: map(_engineer_days, x)},
    {'c_name_feat': [ColumnNames.FEAT_TIME_BIN], 'c_name': [ColumnNamesRaw.CALC_TIME], 'f': _engineer_time_bins}
]
