from sklearn.linear_model import ElasticNet
from sklearn.preprocessing import StandardScaler
import numpy as np

from ..base_model import BaseModel


class ElasticNetWrapper(ElasticNet, BaseModel):
    """
    Class - wrapper for ElasticNet.
    """

    name='ElasticNet'

    def __init__(self, **kwargs):
        """
        Constructor.
        :param dict kwargs:
        """
        super().__init__(**kwargs)
        self.scaler = StandardScaler()

    def fit(self, X, y, sample_weigth=None):
        """
        Method for fitting of the estimator.

        :param numpy.ndarray X: Matrix of input features.
        :param numpy.ndarray y: Vector of ground truths for X.
        :param numpy.ndarray sample_weigth:
        :return: object - self
        """
        self._get_descriptors(X)

        self.scaler = self.scaler.fit(X)
        X = self.scaler.transform(X)

        self.min_y_train = np.percentile(y, 10)
        self.max_y_train = max(y)

        # ElasticNet requires Fortran style arrays
        X = np.asfortranarray(X)
        y = np.asfortranarray(y)
        super().fit(X, y)
        return self

    def predict(self, X):
        """
        Method for prediction.

        :param numpy.ndarray X: Data, which should be used as input for the estimation of predicted value.
        :return: np.ndarray
        """
        X = self._impute_prediction_sample(X=X)
        X = self.scaler.transform(X)
        X = np.asfortranarray(X)
        pred = super().predict(X=X)
        return self._coerce(pred)

    def dump(self):
        """
        Method for dumping of the model providing descriptors of model in dict.

        :return: dict
        """
        d = dict()
        d['model'] = dict()
        d['model']['coef_'] = self.coef_.tolist()
        d['model']['intercept_'] = self.intercept_.tolist()
        d['model']['n_iter_'] = self.n_iter_

        d['model']['min_y_train'] = self.min_y_train
        d['model']['max_y_train'] = self.max_y_train

        d['model']['scaler'] = dict()
        d['model']['scaler']['scale_'] = self.scaler.scale_.tolist()
        d['model']['scaler']['mean_'] = self.scaler.mean_.tolist()
        d['model']['scaler']['var_'] = self.scaler.var_.tolist()
        d['model']['scaler']['n_samples_seen_'] = self.scaler.n_samples_seen_

        d['model']['_median_imputer'] = self._median_imputer.tolist()

        d['model_type'] = self.name
        return d

    @staticmethod
    def load(model):
        """
        Method for loading of the model - recreating it from dumped dict.

        :param dict model: dumped model
        :return: ElasticNetWrapper
        """
        inst = ElasticNetWrapper()

        inst.coef_ = np.array(model['coef_'])
        inst.intercept_ = np.array(model['intercept_'])
        inst.n_iter_ = model['n_iter_']

        inst.min_y_train = model['min_y_train']
        inst.max_y_train = model['max_y_train']

        inst.scaler.scale_ = np.array(model['scaler']['scale_'])
        inst.scaler.mean_ = np.array(model['scaler']['mean_'])
        inst.scaler.var_ = np.array(model['scaler']['var_'])
        inst.scaler.n_samples_seen_d = model['scaler']['n_samples_seen_']

        inst._median_imputer = np.array(model['_median_imputer'])

        return inst