from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
import numpy as np

from ..base_model import BaseModel


class LinearRegressionWrapper(LinearRegression, BaseModel):
    """
    Class - wrapper for LinearRegression.
    """

    name='LinearRegression'

    def __init__(self, **kwargs):
        """
        Constructor.

        :param dict kwargs:
        """
        super().__init__(**kwargs)
        self.scaler = StandardScaler()

    def fit(self, X, y, sample_weight=None):
        """
        Method for fitting of the estimator.

        :param numpy.ndarray X:
        :param numpy.ndarray y:
        :param numpy.ndarray sample_weight:
        :return:
        """
        self._get_descriptors(X)

        self.scaler = self.scaler.fit(X)
        X = self.scaler.transform(X)

        self.min_y_train = np.percentile(y, 10)
        self.max_y_train = 1.5*max(y)

        super().fit(X, y, sample_weight)
        return self

    def predict(self, X):
        """
        Method for the prediction of the output values given the feature vector.

        :param numpy.ndarray X:
        :return: list
        """
        X = self._impute_prediction_sample(X=X)
        X = self.scaler.transform(X)
        pred = super().predict(X=X)
        return self._coerce(pred)

    def dump(self):
        """
        Method for dumping of the existing model.

        :return: dict
        """
        d = dict()
        d['model'] = dict()
        d['model']['coef_'] = self.coef_.tolist()
        d['model']['intercept_'] = self.intercept_.tolist()

        d['model']['min_y_train'] = self.min_y_train
        d['model']['max_y_train'] = self.max_y_train

        d['model']['scaler'] = dict()
        d['model']['scaler']['scale_'] = self.scaler.scale_.tolist()
        d['model']['scaler']['mean_'] = self.scaler.mean_.tolist()
        d['model']['scaler']['var_'] = self.scaler.var_.tolist()
        d['model']['scaler']['n_samples_seen_'] = self.scaler.n_samples_seen_

        d['model']['_median_imputer'] = self._median_imputer.tolist()

        d['model_type'] = self.name
        return d

    @staticmethod
    def load(model):
        """
        Method for recreating model from dumped data.

        :param dict model:
        :return: LinearRegressionWrapper
        """
        inst = LinearRegressionWrapper()

        inst.coef_ = np.array(model['coef_'])
        inst.intercept_ = np.array(model['intercept_'])

        inst.min_y_train = model['min_y_train']
        inst.max_y_train = model['max_y_train']

        inst.scaler.scale_ = np.array(model['scaler']['scale_'])
        inst.scaler.mean_ = np.array(model['scaler']['mean_'])
        inst.scaler.var_ = np.array(model['scaler']['var_'])
        inst.scaler.n_samples_seen_d = model['scaler']['n_samples_seen_']

        inst._median_imputer = np.array(model['_median_imputer'])

        return inst