from sklearn.linear_model import Ridge
from sklearn.preprocessing import StandardScaler
import numpy as np

from ..base_model import BaseModel


class RidgeWrapper(Ridge, BaseModel):
    """
    Class - wrapper for RidgeRegression.
    """

    name='Ridge'

    def __init__(self, **kwargs):
        """
        Constructor.

        :param dict kwargs:
        """
        super().__init__(**kwargs)
        self.scaler = StandardScaler()

    def fit(self, X, y, sample_weight=None):
        """
        Method for prediction.

        :param numpy.ndarray X:
        :param numpy.ndarray y:
        :param numpy.ndarray sample_weight:
        :return: object - self
        """
        self._get_descriptors(X)

        self.scaler = self.scaler.fit(X)
        X = self.scaler.transform(X)

        self.min_y_train = np.percentile(y, 10)
        super().fit(X, y, sample_weight)
        return self

    def predict(self, X):
        """
        Method for prediction.

        :param numpy.ndarray X:
        :return: list
        """
        X = self._impute_prediction_sample(X=X)
        X = self.scaler.transform(X)
        pred = super().predict(X=X)
        return self._coerce(pred)

    def dump(self):
        """
        Method for dumping of the estimator.

        :return: dict
        """
        d = dict()
        d['model'] = dict()
        d['model']['coef_'] = self.coef_.tolist()
        d['model']['intercept_'] = self.intercept_.tolist()
        d['model']['n_iter_'] = self.n_iter_.tolist() if self.n_iter_ is not None else ""

        d['model']['min_y_train'] = self.min_y_train

        d['model']['scaler'] = dict()
        d['model']['scaler']['scale_'] = self.scaler.scale_.tolist()
        d['model']['scaler']['mean_'] = self.scaler.mean_.tolist()
        d['model']['scaler']['var_'] = self.scaler.var_.tolist()
        d['model']['scaler']['n_samples_seen_'] = self.scaler.n_samples_seen_

        d['model']['_median_imputer'] = self._median_imputer.tolist()

        d['model_type'] = self.name
        return d

    @staticmethod
    def load(model):
        """
        Method for loading of the estimator - recreating it from dumped dict.

        :param dict model:
        :return: instance of RidgeWrapper
        """
        inst = RidgeWrapper()

        inst.coef_ = np.array(model['coef_'])
        inst.intercept_ = np.array(model['intercept_'])
        inst.n_iter_ = np.array(model['n_iter_']) if model['n_iter_'] != "" else None

        inst.min_y_train = model['min_y_train']

        inst.scaler.scale_ = np.array(model['scaler']['scale_'])
        inst.scaler.mean_ = np.array(model['scaler']['mean_'])
        inst.scaler.var_ = np.array(model['scaler']['var_'])
        inst.scaler.n_samples_seen_d = model['scaler']['n_samples_seen_']

        inst._median_imputer = np.array(model['_median_imputer'])

        return inst